@@ -1,10 +1,10 @@
 ;;; setup.el --- Helpful Configuration Macro    -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021  Free Software Foundation, Inc.
+;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
 
 ;; Author: Philip Kaludercic <philipk@posteo.net>
-;; Maintainer: Philip Kaludercic <philipk@posteo.net>
-;; Version: 1.2.0
+;; Maintainer: Philip Kaludercic <~pkal/public-inbox@lists.sr.ht>
+;; Version: 1.3.0
 ;; Package-Requires: ((emacs "26.1"))
 ;; Keywords: lisp, local
 ;; URL: https://git.sr.ht/~pkal/setup
@@ -37,8 +37,15 @@
 
 ;;; News:
 
+;;;; Version 1.3.0
+
+;; - Add new `:and' macro.
+;; - Add new `:with-function' macro and have `:hook-into' use
+;;   the context function instead of the context mode (Earl Hyatt)
+;; - Improve `:bind-into' handling of maps.
+
 ;;;; Version 1.2.0
-;;
+
 ;; - Remove `setup-wrap-to-demote-errors' from `setup-modifier-list'
 ;; - Pull `setup-expand-local-macros'  back into `setup'
 ;; - Let `:with-feature' and `:with-mode' check symbol properties to
@@ -50,9 +57,9 @@
 ;; - Add :ensure key to `setup-define' to replace
 ;;   the deprecated setup-ensure-... functions
 ;; - Add `setup-bind' macro to simplify context modification.
-;;
+
 ;;;; Version 1.1.0:
-;;
+
 ;; - Fix quoting error in :file-match definition
 ;; - Remove unnecessary check for lexical binding
 ;; - Change `setup-define' indentation
@@ -107,7 +114,7 @@ Do not modify this variable by hand.  Instead use
 `setup-define.'")
 
 ;;;###autoload
-(defun setup-make-docstring ()
+(defun setup--make-docstring ()
   "Return a docstring for `setup'."
   (with-temp-buffer
     (insert (documentation (symbol-function 'setup) 'raw)
@@ -154,13 +161,13 @@ NAME may also be a macro, if it can provide a symbol."
                 (if (assq :with-feature setup-macros)
                     `(:with-feature ,name ,@body)
                   (macroexp-progn body))
-                setup-macros))
+                (append setup-macros macroexpand-all-environment)))
     (dolist (mod-fn setup-modifier-list)
       (setq body (funcall mod-fn body name)))
     body))
 
 ;;;###autoload
-(put 'setup 'function-documentation '(setup-make-docstring))
+(put 'setup 'function-documentation '(setup--make-docstring))
 
 (defun setup-define (name fn &rest opts)
   "Define `setup'-local macro NAME using function FN.
@@ -272,14 +279,14 @@ functions `func'.  Any other value is invalid."
                  (cons '&rest spec))
                 (t spec)))))
 
-(defun setup-xref-def-function (symbol)
+(defun setup--xref-def-function (symbol)
   "Return an elisp xref location for SYMBOL."
   (and (assq symbol setup-macros)
        (let ((file (get symbol 'setup-definition-file)))
          (list (elisp--xref-make-xref nil symbol file)))))
 
 (add-to-list 'elisp-xref-find-def-functions
-             #'setup-xref-def-function)
+             #'setup--xref-def-function)
 
 
 ;;; Common utility functions for local macros
@@ -366,6 +373,7 @@ VAL into one s-expression."
                     (setup-bind body
                       (feature feature)
                       (mode (or (get features 'setup-mode) mode))
+                      (func (or (get features 'setup-func) mode))
                       (hook (or (get features 'setup-hook)
                                 (get mode 'setup-hook)
                                 (intern (format "%s-hook" mode))))
@@ -376,10 +384,14 @@ VAL into one s-expression."
               bodies))
       (macroexp-progn (if features (nreverse bodies) body))))
   :documentation "Change the FEATURE that BODY is configuring.
-This macro also declares a current mode by appending \"-mode\" to
-FEATURE, unless it already ends with \"-mode\".
+This macro also:
+- Declares a current mode by appending \"-mode\" to
+  FEATURE, unless it already ends with \"-mode\"
+- Declares a current hook by appending \"-hook\" to the mode
+- Declares a current map by appending \"-map\" to the mode
+- Declares a current function that has the same name as the mode
 If FEATURE is a list, apply BODY to all elements of FEATURE."
-  :debug '(sexp setup)
+  :debug '([&or ([&rest sexp]) sexp] setup)
   :indent 1)
 
 (setup-define :with-mode
@@ -388,6 +400,8 @@ If FEATURE is a list, apply BODY to all elements of FEATURE."
       (dolist (mode (if (listp modes) modes (list modes)))
         (push (setup-bind body
                 (mode mode)
+                (func (or (get mode 'setup-func)
+                          mode))
                 (hook (or (get mode 'setup-hook)
                           (intern (format "%s-hook" mode))))
                 (map (or (get mode 'setup-map)
@@ -395,8 +409,12 @@ If FEATURE is a list, apply BODY to all elements of FEATURE."
               bodies))
       (macroexp-progn (nreverse bodies))))
   :documentation "Change the MODE that BODY is configuring.
-If MODE is a list, apply BODY to all elements of MODE."
-  :debug '(sexp setup)
+If MODE is a list, apply BODY to all elements of MODE.
+This macro also:
+- Declares a current hook by appending \"-hook\" to the mode
+- Declares a current map by appending \"-map\" to the mode
+- Declares a current function that has the same name as the mode"
+  :debug '([&or ([&rest sexp]) sexp] setup)
   :indent 1)
 
 (setup-define :with-map
@@ -408,7 +426,7 @@ If MODE is a list, apply BODY to all elements of MODE."
       (macroexp-progn (nreverse bodies))))
   :documentation "Change the MAP that BODY will bind to.
 If MAP is a list, apply BODY to all elements of MAP."
-  :debug '(sexp setup)
+  :debug '([&or ([&rest sexp]) sexp] setup)
   :indent 1)
 
 (setup-define :with-hook
@@ -420,6 +438,21 @@ If MAP is a list, apply BODY to all elements of MAP."
       (macroexp-progn (nreverse bodies))))
   :documentation "Change the HOOK that BODY will use.
 If HOOK is a list, apply BODY to all elements of HOOK."
+  :debug '([&or ([&rest sexp]) sexp] setup)
+  :indent 1)
+
+(setup-define :with-function
+  (lambda (functions &rest body)
+    (let (bodies)
+      (dolist (func (if (listp functions) functions (list functions)))
+        (let ((fn (if (memq (car-safe func) '(quote function))
+                      (cadr func)
+                    func)))
+          (push (setup-bind body (func fn))
+                bodies)))
+      (macroexp-progn (nreverse bodies))))
+  :documentation "Change the FUNCTION that BODY will use.
+If FUNCTION is a list, apply BODY to all elements of FUNCTION."
   :debug '(sexp setup)
   :indent 1)
 
@@ -430,8 +463,7 @@ If HOOK is a list, apply BODY to all elements of HOOK."
          (package-refresh-contents))
        (package-install ',package)))
   :documentation "Install PACKAGE if it hasn't been installed yet.
-This macro can be used as NAME, and it will replace itself with
-the first PACKAGE."
+The first PACKAGE can be used to deduce the feature context."
   :repeatable t
   :shorthand #'cadr)
 
@@ -440,8 +472,7 @@ the first PACKAGE."
     `(unless (require ',feature nil t)
        ,(setup-quit)))
   :documentation "Try to require FEATURE, or stop evaluating body.
-This macro can be used as NAME, and it will replace itself with
-the first FEATURE."
+The first FEATURE can be used to deduce the feature context."
   :repeatable t
   :shorthand #'cadr)
 
@@ -484,9 +515,11 @@ the first FEATURE."
   :repeatable t)
 
 (setup-define :bind-into
-  (lambda (feature &rest rest)
-    `(:with-feature ,feature (:bind ,@rest)))
-  :documentation "Bind into keys into the map of FEATURE.
+  (lambda (feature-or-map &rest rest)
+    (if (string-match-p "-map\\'" (symbol-name feature-or-map))
+        `(:with-map ,feature-or-map (:bind ,@rest))
+      `(:with-feature ,feature-or-map (:bind ,@rest))))
+  :documentation "Bind into keys into the map of FEATURE-OR-MAP.
 The arguments REST are handled as by `:bind'."
   :debug '(sexp &rest form sexp)
   :indent 1)
@@ -504,8 +537,8 @@ The arguments REST are handled as by `:bind'."
                    (if (string-match-p "-hook\\'" name)
                        mode
                      (intern (concat name "-hook"))))
-               #',(setup-get 'mode)))
-  :documentation "Add current mode to HOOK."
+               #',(setup-get 'func)))
+  :documentation "Add current function to HOOK."
   :repeatable t)
 
 (setup-define :option
@@ -588,8 +621,7 @@ supported:
     `(unless (package-installed-p ',package)
        ,(setup-quit)))
   :documentation "If package is not installed, stop evaluating the body.
-This macro can be used as NAME, and it will replace itself with
-the first PACKAGE."
+The first PACKAGE can be used to deduce the feature context."
   :repeatable t
   :shorthand #'cadr)
 
@@ -598,8 +630,7 @@ the first PACKAGE."
     `(unless (featurep ',feature)
        ,(setup-quit)))
   :documentation "If FEATURE is not available, stop evaluating the body.
-This macro can be used as NAME, and it will replace itself with
-the first PACKAGE."
+The first FEATURE can be used to deduce the feature context."
   :repeatable t
   :shorthand #'cadr)
 
@@ -629,6 +660,22 @@ yourself."
   :after-loaded t
   :indent 0)
 
+(setup-define :and
+  (lambda (&rest conds)
+    `(if (and ,@(butlast conds))
+         ,@(last conds)
+       ,(setup-quit)))
+  :documentation "Abort evaluation of CONDS are not all true.
+The expression of the last condition is used to deduce the
+feature context."
+  :shorthand
+  (lambda (head)
+    (unless (cdr head)
+      (error ":and requires at least one condition"))
+    (let ((shorthand (get (caar (last head)) 'setup-shorthand)))
+      (and shorthand (funcall shorthand (car (last head))))))
+  :debug '(setup))
+
 
 ;;; Obsoleted code
 
